from AhpAnpLib import functions_AHPLib as reqLib
from AhpAnpLib import ratings_AHPLib as rate
# import functions_AHPLib as reqLib
# import ratings_AHPLib as rate
class Model:
    id_generator = reqLib.itertools.count(0)
    def __init__(self,name=None):
        
        # choose appropriate comparative word that will be saved in the keyword variable. Usually we use: "important","preferred", "likely" can support different keywords
        self.modelID=next(self.id_generator)
        if(name!=None):
            self.name=name
        else:
            self.name="Untitled"
        self.clusters=[]
        # modelType pairwise or ratings 
        self.modelType="pairwise" 
        self.nodeConnections=[]
        self.clusterConnections=[]

        self.all_pc_matrices=[]
        self.all_pr_vectors=[]
        self.pc_with_respect_to=[]
        self.pc_node_order=[]
        self.cluster_neighbors_mtrx=[]
        self.cluster_pc_matrices=[]
        self.supermatrix=[]
        self.weighted_supermatrix=[]
        self.limit=[]
        self.questionnaires=[]
        self.nlabels=[]
        self.clabels=[]
        self.wrtlabels=[]
        self.rateModel=rate.RatStruct(self)
        

        

    def __repr__(self):
        return "\nModel Name: "+self.name+"\nType:"+self.modelType+"\nNodes: "+str(self.getNodeListFromClusters())+"\n Clusters: "+str(self.clusters)+"\n Node Connections: "+str(self.nodeConnections)+"\n Cluster Connections: "+str(self.clusterConnections)
    
    def addCluster2Model(self,cluster):
    #add a cluster that DOES NOT belong to a model to this model
        if isinstance(cluster, (Cluster)):
            if (cluster.parModel==""):
                if cluster not in self.clusters:
                    self.clusters.append(cluster)
                    cluster.parModel=self.modelID
                    # print(cluster.parModel, "--",self.modelID)
            else:
                print("Cluster already assigned to model")
        else:
            print(f"Error: Given object {cluster} not a cluster")
    def addMultipleClusters2Model(self,*clusters):
         for cluster in clusters:
             self.addCluster2Model(cluster)
    
    def remClusterByNameFromModel(self,clusterName):
        cluster=self.getClusterObjByName(clusterName)
    #add a cluster that DOES NOT belong to a model to this model
        # print(cluster.parModel, "--",self.modelID)
        if (cluster.parModel==self.modelID):
            if cluster in self.clusters:
                self.clusters.remove(cluster)
                cluster.parCluster=""

                #   update model connections
                self.defineAllNodeConnections()
                #update cluster connections
                self.defineAllClusterConnections()
        else:
            print("Cluster not assigned to this model")

    def remClusterObjFromModel(self,clusterObj):
        # print(cluster.parModel, "--",self.modelID)
        if (clusterObj.parModel==self.modelID):
            if clusterObj in self.clusters:
                self.clusters.remove(clusterObj)
                clusterObj.parCluster=""

                #   update model connections
                self.defineAllNodeConnections()
                #update cluster connections
                self.defineAllClusterConnections()
        else:
            print("Cluster not assigned to this model")

    def getClusterListFromClusters(self):
        clusters=[]
        for cluster in sorted(self.clusters, key=reqLib.op.attrgetter('order')):
            for node in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')):
                parCluster=node.parCluster
                clusters.append(parCluster)
        return clusters
    def getListNumNodesInClusters(self):
        clusters=[]
        for cluster in sorted(self.clusters, key=reqLib.op.attrgetter('order')):
            tot=0
            for node in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')):
                tot+=1
            clusters.append(tot)
        return clusters
    def getNodeListFromClusters(self):
        nodes=[]
        for cluster in sorted(self.clusters, key=reqLib.op.attrgetter('order')):
            for node in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')):
                nodes.append(node)
        return nodes
    def defineAllNodeConnections(self):
        self.nodeConnections={}
        for cluster in self.clusters:
            for node in cluster.nodes:
                self.nodeConnections[node]=node.connectedTo
        return self.nodeConnections
    
    def getClusterIndexByName(self,name):
        for index, item in enumerate(self.clusters):
            if item.name == name:
                break
        else:
            index = -1
        return index
    
    def getClusterObjByName(self,name):
        for item in self.clusters:
            # print(item.name)
            if item.name == name:
                return item
        return -1
        
    def getClusterObjByID(self,ID):
        for item in self.clusters:
            # print(item.name)
            if item.clusterID == ID:
                return item
        return -1
    
    def getClusterIDByName(self,name):
        for item in self.clusters:
            # print(item.name)
            if item.name == name:
                return item.clusterID
        return -1

    def getNodeIndexInClusterModelByName(self,name):
        indexN = -1
        for index, item in enumerate(self.clusters):
            for indexN, itemN in enumerate(item.nodes):
                if itemN.name == name:
                    return indexN
        return indexN
    
    def getNodeInClusterModelByName(self,name):

        for item in self.clusters:
            for itemN in item.nodes:
                if itemN.name == name:
                    # print(itemN)
                    return itemN
        return -1
    
    def getNodeIDByName(self,name):

        for item in self.clusters:
            for itemN in item.nodes:
                if itemN.name == name:
                    # print(itemN)
                    return itemN.nodeID
        return -1

    def getNodeObjByName(self,name):
        for item in self.clusters:
            for itemN in item.nodes:
                if itemN.name == name:
                    # print(itemN)
                    return itemN
        return -1
    
    def addNodeConnectionFromTo(self,nodeFromName,nodeToName, verb=False):
         #will add a connection from nodeFrom to the nodeTo node

        nodeFrom=self.getNodeInClusterModelByName(nodeFromName)
        # print("From {} ".format(nodeFrom))
        nodeTo=self.getNodeInClusterModelByName(nodeToName)
        # print("To {} ".format(nodeTo))
        if (isinstance(nodeFrom, (Node)) & isinstance(nodeTo, (Node))) :
            if nodeTo not in nodeFrom.connectedTo:
                nodeFrom.connectedTo.append(nodeTo)
            else:
                print("Trying to add n{}. Node already in connections.".format(str(nodeTo.nodeID)))
            if verb==True: 
                print("node: "+nodeFrom.name+" connectedTo: "+str(nodeFrom.connectedTo))

            #   update model connections
            self.defineAllNodeConnections()
            #update cluster connections
            self.defineAllClusterConnections()
        else:
            print("Error: Nodes not found or given objects not nodes in addNodeConnectionFromTo")


    def addNodeConnectionFromNodeToAllNodesOfCluster(self,nodeFromName,clusterToName):
         #will add a connection from nodeFrom to each nodeTo of clusterToName
         clusterTo=self.getClusterObjByName(clusterToName)
         for nodeTo in clusterTo.nodes:
            self.addNodeConnectionFromTo(nodeFromName,nodeTo.name)

    def addNodeConnectionFromAllNodesToAllNodesOfCluster(self,clusterFromName,clusterToName):
         #will add a connection from each nodeFrom of clusterFromName to each nodeTo of clusterToName
        clusterFrom=self.getClusterObjByName(clusterFromName)
        if isinstance(clusterFrom, (Cluster)):
            for nodeFrom in clusterFrom.nodes:
                self.addNodeConnectionFromNodeToAllNodesOfCluster(nodeFrom.name,clusterToName)   
        else:
            print(f"Error: Given object not a cluster in addNodeConnectionFromAllNodesToAllNodesOfCluster")
    def addNodeConnectionFromNodeToNodeList(self,nodeFromName,*nodeNames):
         #will add a connection from  nodeFrom to each nodeTo
        nodeFromObj=self.getNodeInClusterModelByName(nodeFromName)
        if isinstance(nodeFromObj, (Node)):
            for nodeToName in nodeNames:
                nodeToTempObj=self.getNodeInClusterModelByName(nodeToName)
                if isinstance(nodeToTempObj, (Node)):
                    self.addNodeConnectionFromTo(nodeFromName,nodeToName)
                else:
                    print(f"Error: Given object {nodeToTempObj} not a node in addNodeConnectionFromNodeToNodeList")  
        else:
            print(f"Error: Given object {nodeFromObj} not a node in addNodeConnectionFromNodeToNodeList")
    def remNodeConnectionFromTo(self,nodeFromName,nodeToName,verb=False):
        flag=0
        nodeFrom=self.getNodeInClusterModelByName(nodeFromName)
        # print("From {} ".format(nodeFrom))
        nodeTo=self.getNodeInClusterModelByName(nodeToName)
        # print("To {} ".format(nodeTo))

        if nodeTo in nodeFrom.connectedTo:
            nodeFrom.connectedTo.remove(nodeTo)
        else:
            print("Trying to delete n{}. Node not in connections.".format(str(nodeTo.nodeID)))
        if verb==True: 
            print("node: "+nodeFrom.name+" after the deletion is now connected to: "+str(nodeFrom.connectedTo))

        #   update model connections
        self.defineAllNodeConnections()
        #update cluster connections
        self.defineAllClusterConnections()
   

    def defineAllClusterConnections(self):
        self.clusterConnections={}
        for cluster in self.clusters:
            self.clusterConnections[cluster]=cluster.calcAllConnectionsFrom()
        return self.clusterConnections

    def remAllNodeConnectionsTo(self,nodeName):
        flag=0
        nodeTo_del_id=self.getNodeIDByName(nodeName)
        for nodeFrom in self.getNodeListFromClusters():
            for nodeTo in nodeFrom.connectedTo:
                if nodeTo.nodeID==nodeTo_del_id:
                    flag=1
                    nodeFrom.connectedTo.remove(nodeTo)
                    print("node connection from: "+nodeFrom.name+" to :"+nodeName+" removed.")
                    #update cluster connections
                    nodeTo.parCluster.calcAllConnectionsFrom()
        if flag==0:
            print("Trying to delete n{}. Node was not found.".format(nodeName))
      

    def remAllNodeConnectionsFrom(self,nodeFromName):
        flag=0 
        toDel=[]
        nodeFrom=self.getNodeInClusterModelByName(nodeFromName)
        if(nodeFrom==-1):
            print("Node not found in model - process could not be completed")
        else:
            print(f"Connected to: {nodeFrom.connectedTo}")
            for nodeTo in nodeFrom.connectedTo:
                toDel.append(nodeTo)
            print(*toDel, sep='\n')
            for nodeTo in toDel:
                self.remNodeConnectionFromTo(nodeFrom.name,nodeTo.name)


    def showAllNodeConnectionsFrom(self, nodeFromName):
        flag=0
        nodeFrom_id=self.getNodeIDByName(nodeFromName)
        for nodeFrom in self.getNodeListFromClusters():
            if nodeFrom.nodeID==nodeFrom_id:
                
                for nodeTo in nodeFrom.connectedTo:
                    if flag==0:
                        print("Connections from node "+str(nodeFrom) +" to: ")
                    flag=1
                    print(str(nodeTo))
        if(flag==0):
            print("No connections from",nodeFromName)
    def retAllNodeConnectionsFrom(self, nodeFromName):
        nodelist_connected_to=[]
        for nodeFrom in self.getNodeListFromClusters():
            if(nodeFrom.name==nodeFromName):
                for nodeTo in nodeFrom.connectedTo:
                        nodelist_connected_to.append(nodeTo)
        return nodelist_connected_to



    def showAllNodeConnectionsTo(self, nodeToName):
        #same with enumerating the connectedTo list of each node
        flag=0
        nodeTo_id=self.getNodeIDByName(nodeToName)
        for nodeFrom in self.getNodeListFromClusters():
            for nodeTo in nodeFrom.connectedTo:
                if nodeTo.nodeID==nodeTo_id:
                    flag=1
                    print("Connections from node "+str(nodeFrom) +" to: "+str(nodeTo))        

        if(flag==0):
            print("No connections to",nodeToName)

    def showAllNodeConnections(self):
        #same with enumerating the connectedTo list of each node
        for nodeFrom in self.getNodeListFromClusters():
            flag=0
            for nodeTo in nodeFrom.connectedTo:
                    if(flag==0):
                        print("Connections from node "+str(nodeFrom))
                    flag=1
                    print(" to: "+str(nodeTo))
            if(flag==0):
                print("No connections from",nodeFrom.name)

    def showAllClusterConnections(self):
        #same with enumerating the connectedTo list of each node
        for clusterFrom in self.clusters:
            flag=0
            for clusterTo in clusterFrom.connectedTo:
                    if(flag==0):
                        print("Connection(s) from cluster "+str(clusterFrom))
                    flag=1
                    print(" to: "+str(clusterTo))
            if(flag==0):
                print("No connections from",clusterFrom.name)  

    def retAllClusterConnectionsFrom(self,clusterName,verb=False):
        #same with enumerating the connectedTo list of each node
        connected_to=[]
        clusterFrom_id=self.getClusterIDByName(clusterName)
        for clusterFrom in self.clusters:
            flag=0
            if clusterFrom.clusterID==clusterFrom_id:
                if(flag==0):
                    if(verb):
                        print("Connection(s) from cluster "+str(clusterFrom.name))
                    flag=1
                inner_flag=0
                for clusterTo in clusterFrom.connectedTo:
                    connected_to.append(clusterTo)
                    inner_flag=1
                    if(verb):
                        print(" to: "+str(clusterTo.name))
                if(inner_flag==0):
                    if(verb):
                        print("No connections from",clusterFrom.name)
                    return None
                     
        if(verb):       
            print(f"Results list:{connected_to}")
        return connected_to 

    def retAllClusterConnectionsFromNode(self,nodeName,verb=False):
        #to which clusters is this node connected to
        connected_to=[]
        flag=0
        node_id=self.getNodeIDByName(nodeName)

        # print("working with",node_id)
        for nodeFrom in self.getNodeListFromClusters():
            if nodeFrom.nodeID==node_id:
                for nodeTo in nodeFrom.connectedTo:
                    flag=1
                    clusterTo=nodeTo.parCluster
                    if(clusterTo not in connected_to):
                        if (verb):
                            print("Connections from node "+str(nodeFrom.name) +" to cluster: "+str(clusterTo.name)) 
                        connected_to.append(clusterTo)      
        if(flag==0):
             if verb==True:
                print("No connections from",nodeName)
        return connected_to
    
    def showAllClusterConnectionsFrom(self,clusterName):
        #same with enumerating the connectedTo list of each node
        clusterFrom_id=self.getClusterIDByName(clusterName)
        for clusterFrom in self.clusters:
            flag=0
            if clusterFrom.clusterID==clusterFrom_id:
                for clusterTo in clusterFrom.connectedTo:
                        if(flag==0):
                            print("Connection(s) from cluster "+str(clusterFrom))
                            flag=1
                        print(" to: "+str(clusterTo))
                if(flag==0):
                    print("No connections from",clusterFrom.name) 
        
    def listAllClusterConnections(self,clusterFrom):
        connected_to=[]
        #same with enumerating the connectedTo list of each node
        
        for clusterTo in clusterFrom.connectedTo:
            connected_to.append(clusterTo)
        return connected_to       
    
   
       
    def moveNode2Cluster(self,node,toCluster,withConnections,verb=False):    
        #moving node from Cluster A to Cluster B (toCluster) of a specific model
        my_parCluster=node.parCluster
        parID=self.getClusterObjByID(my_parCluster.clusterID)
        if(verb):
            print(f"Initially we have parent cluster {my_parCluster.name} of node {node.name}")
        if(parID!=-1):
            my_cl_indx=self.getClusterIndexByName(my_parCluster.name)
            if(my_cl_indx!=-1):
                self.clusters[my_cl_indx].nodes.remove(node)
                if(verb):
                    print(f"Now my cluster has the following nodes:{self.clusters[my_cl_indx].nodes}")
            dest_cl_indx=self.getClusterIndexByName(toCluster.name)
            if(dest_cl_indx!=-1):
                self.clusters[dest_cl_indx].nodes.append(node)
                self.clusters[dest_cl_indx].nodes[-1].parCluster=dest_cl_indx
                if(withConnections==False):
                    self.clusters[dest_cl_indx].nodes[-1].connectedTo=[]
                    #update cluster connections
                self.defineAllClusterConnections()
                if(verb):
                    print(f"My destination cluster has the following nodes:{self.clusters[dest_cl_indx].nodes}")
                    print(f"My node's parent now is :{self.clusters[dest_cl_indx].nodes[-1].parCluster}")
                    print(f"My node's connections To now are :{self.clusters[dest_cl_indx].nodes[-1].connectedTo}")
    def getMaxNodeOrder(self):
        max=-1
        for cluster in self.clusters:
            for node in cluster.nodes:
             if(node.order>max):
                 max=node.order
        return max
    def getMaxClusterOrder(self):
        max=-1
        for cluster in self.clusters:
            if(cluster.order>max):
                max=cluster.order
        return max
    def printStruct(self):
        print("_________________________MODEL STRUCTURE_________________________")
        
        print("Name: "+self.name)
        print("Type: "+self.modelType)
        print("\n____________________________NODES_______________________________")
        for this_node in self.getNodeListFromClusters():
            print(this_node)
        print("____________________________CLUSTERS____________________________")
        for this_cluster in self.clusters:
            print(this_cluster)
        print("_________________________NODE CONNECTIONS___________________________")
        self.showAllNodeConnections()
        print("_________________________CLUSTER CONNECTIONS__________________")
        self.showAllClusterConnections()
        if(self.modelType=="ratings"):
            print("_________________________RATING MODEL SET UP__________________")
            print(self.rateModel)

    def setModelTypeRatings(self):
        self.modelType="ratings"
    def drawGraphNodes(self):
        G = reqLib.nx.DiGraph()

        # Add nodes with cluster assignments
        node_colors = []
        color_mapping = {}  # Dictionary to store color assignments
        for clusterObj in self.clusters:
            for nodeObj in clusterObj.nodes:
                cluster_id = nodeObj.parCluster.clusterID
                G.add_node(nodeObj, cluster=cluster_id, name=nodeObj.name)
                if cluster_id not in color_mapping:
                # Assign a new color for the cluster
                    color_mapping[cluster_id] = len(color_mapping)
                node_colors.append(color_mapping[cluster_id])
        # print(f"colors{color_mapping} nodes {node_colors}")

        # Add edges 
        for clusterObj in self.clusters:
            for nodeObj in clusterObj.nodes:
                nodes_connected_to=self.retAllNodeConnectionsFrom(nodeObj.name)
                # print(f"for {nodeObj} we have {nodes_connected_to}")
                for node_to in nodes_connected_to:
                    if(len(nodes_connected_to)>0):
                        # print(f"From {nodeObj.name} to {node_to.name}")
                        G.add_edge(nodeObj, node_to) 

        # Generate the positions of the nodes
        pos = reqLib.nx.circular_layout(G)
        # pos = reqLib.nx.shell_layout(G)
        
        # Set the size of the initial graph
        reqLib.plt.figure(figsize=(18, 9)) 
        
        # Draw nodes with colors based on clusters
        node_size=8000
        nodes=reqLib.nx.draw_networkx_nodes(G, 
                                            pos, 
                                            node_color=node_colors, 
                                            cmap="rainbow", 
                                            alpha=0.5,
                                            node_shape=">",
                                            node_size=node_size)

        # Draw edges (if you have edges in your graph)
        # reqLib.nx.draw_networkx_edges(G, pos)
        reqLib.nx.draw_networkx_edges(
        G,
        pos,
        # connectionstyle="arc3,rad=0.1",  # Add this parameter to create curved edges
        arrows=True,  # Add arrows to the edges
        )
        # Add labels to nodes
        labels = reqLib.nx.get_node_attributes(G, "name")
        reqLib.nx.draw_networkx_labels(G, pos, labels)

        # Create a legend
        unique_clusters = sorted(color_mapping.keys())
        legend_handles = []
        legend_labels = []
        color_mapping_values = list(color_mapping.values())
        cmap = reqLib.plt.cm.get_cmap("rainbow", len(color_mapping_values))
        for cluster, color_value in zip(unique_clusters, color_mapping_values):
            my_color = cmap(color_value)
            handle = reqLib.plt.Line2D([], [], marker="o", markersize=10, color=my_color)
            legend_handles.append(handle)
            legend_labels.append(f" {self.getClusterObjByID(cluster).name}")
        reqLib.plt.legend(legend_handles, legend_labels, title="Clusters")
        # Display the graph
        reqLib.plt.axis("off")
        reqLib.plt.show()

    def drawGraphClusters(self):
        G = reqLib.nx.DiGraph()

        # Add clusters 

        for clusterObj in self.clusters:
                G.add_node(clusterObj.name, name=clusterObj.name)
                

        # Add edges 
        for clusterObj in self.clusters:
            clusters_connected_to=self.retAllClusterConnectionsFrom(clusterObj.name)
            print(f"Cluster from {clusterObj.name} connected to: {clusters_connected_to}")
            if(clusters_connected_to!=None):
                    for cluster_to in clusters_connected_to:
                            # print(f"From {nodeObj.name} to {node_to.name}")
                            G.add_edge(clusterObj.name, cluster_to.name) 

        # Generate the positions of the nodes
        pos = reqLib.nx.circular_layout(G)
        # pos = reqLib.nx.shell_layout(G)
        
        # Set the size of the initial graph
        reqLib.plt.figure(figsize=(18, 9)) 
        
        # Draw nodes with colors based on clusters
        node_size=8000
        nodes=reqLib.nx.draw_networkx_nodes(G, 
                                            pos, 
                                            cmap="rainbow", 
                                            alpha=0.5,
                                            node_shape="s",
                                            node_size=node_size)

        # Draw edges (if you have edges in your graph)
        # reqLib.nx.draw_networkx_edges(G, pos)
        reqLib.nx.draw_networkx_edges(
        G,
        pos,
        # connectionstyle="arc3,rad=0.1",  # Add this parameter to create curved edges
        arrows=True,  # Add arrows to the edges
        )
        # Add labels to nodes
        labels = reqLib.nx.get_node_attributes(G, "name")
        reqLib.nx.draw_networkx_labels(G, pos, labels)

        # # Create a legend
        # unique_clusters = sorted(color_mapping.keys())
        # legend_handles = []
        # legend_labels = []
        # color_mapping_values = list(color_mapping.values())
        # cmap = reqLib.plt.cm.get_cmap("rainbow", len(color_mapping_values))
        # for cluster, color_value in zip(unique_clusters, color_mapping_values):
        #     my_color = cmap(color_value)
        #     handle = reqLib.plt.Line2D([], [], marker="o", markersize=10, color=my_color)
        #     legend_handles.append(handle)
        #     legend_labels.append(f" {self.getClusterObjByID(cluster).name}")
        # reqLib.plt.legend(legend_handles, legend_labels, title="Clusters")
        # Display the graph
        reqLib.plt.axis("off")
        reqLib.plt.show()

    def drawGraphModel(self,filepath=None, verb=False):
        
        colors = ['#f57a64', '#64b4f5', '#63c273', '#f5e50c', '#cb7bed', '#f7952d', '#f79ee4', '#9ef7e7']

        # Create the main graph
        g = reqLib.gr.Digraph('G', filename='hierarchy.gv')

        # Create the clusters
        for i,cluster in enumerate(sorted( self.clusters, key=reqLib.op.attrgetter('order'))):
            with g.subgraph(name=f'cluster_{i}') as c:
                c.attr(style='filled', color='lightgrey')
                my_color=colors[i%8]
                for nodeObj in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')):
                    c.node(nodeObj.name, color=my_color, style='filled')  # additional nodes without edges
            c.attr(label=cluster.name)

        

        # Create connections between the nodes in the clusters
   
        # g.edge('a1', 'b3')
        for clusterObj in self.clusters:
            for node_from in clusterObj.nodes:
                nodes_connected_to = self.retAllNodeConnectionsFrom(node_from.name)
                if len(nodes_connected_to) > 0:
                    for node_to in nodes_connected_to:
                        g.edge(node_from.name, node_to.name)

        # Save the graph to a PNG file
        g.format = 'png'
        # g.render(self.name)

        if filepath:
            # Save the graph as a PNG file
            g.render(self.name)

                # Save the PNG image to Excel
            workbook = reqLib.openpyxl.Workbook()
            worksheet = workbook.active
            img = reqLib.Image(self.name + ".png")
            

            # Create a new worksheet named "Graph Model" and insert the PNG image
            worksheet = workbook.create_sheet("Graph Model")
            
            # Save to Excel
            worksheet.add_image(img, "A1")
            workbook.save(filepath)
            if(verb):
                print("PNG image added to the Excel file.")
        else:
            g.render(self.name)
            if(verb):
                print("PNG image saved to your working folder.")



      
class Node:
    id_generator = reqLib.itertools.count(0)

    def __init__(self, node_name,node_order):
        self.name = node_name
        self.order = node_order
        self.nodeID = next(self.id_generator)
        self.parCluster=""
        self.connectedTo=[]

    def __repr__(self):
        return self.name+" order: "+ str(self.order)
        #return self.name+" NID: "+str(self.nodeID)+" order: "+ str(self.order)
    def myParentCluster(self):
        return self.parCluster
    def updateN_DisplayOrder(self,newOrder):
        self.order=newOrder

class Cluster:
    id_generator = reqLib.itertools.count(0)
    
    def __init__(self, cluster_name,cluster_order):
        self.name = cluster_name
        self.order = cluster_order
        self.clusterID = next(self.id_generator)
        self.nodes=[]
        self.connectedTo=[]
        self.parModel=""
       
    def __repr__(self):
        return self.name+" order: "+ str(self.order)
        # return self.name+" CID:  "+str(self.clusterID)+" order: "+ str(self.order)
    def printWithNodes(self):
        print("Cluster: {} with nodes: {}\n".format(self.name,self.nodes))
    def addNode2Cluster(self,node):
    #add a node that DOES NOT belong to a cluster to this cluster
        if isinstance(node, (Node)):
            if (node.parCluster==""):
                if node not in self.nodes:
                    self.nodes.append(node)
                    node.parCluster=self
            else:
                print("Node already assigned to cluster")
        else:
            print(f"Error: Given object {node} not a node")
    def addMultipleNodes2Cluster(self,*nodes):
         for node in nodes:
             self.addNode2Cluster(node)

    # def remNodeFromCluster(self,node):
    # #rem a node from the cluster -- not to be used if node has been assigned to model

    #     # print(self.clusterID, "--",node.parCluster.clusterID,"--",node.parCluster.parModel)
    #     if (node.parCluster.parModel==""):
    #         if (node.parCluster.clusterID==self.clusterID):
    #             if node not in self.nodes:
    #                 self.nodes.remove(node)
    #                 node.parCluster=""
    #         else:
    #             print("Node not assigned to this cluster")
    #     else: 
    #             print("Node already assigned to model - cannot be removed with this command. Remove it from the model by deleting the line corresponding to the assignment of the node to the model ")
    def forceAddNode2Cluster(self,node):
    #add a node even if it already belongs to another cluster to a this cluster
        ClusterA=node.parCluster
        ClusterA.nodes.remove(node)
        if node not in self.nodes:
                self.nodes.append(node)
                node.parCluster=self

    # def detachFromModel(self):
    #     self.parModel=""

    def calcAllConnectionsFrom(self,verb=False):
        # print("cluster connections for: "+self.name)
        # print('cluster nodes: {}'.format(self.nodes))
        for fromNode in self.nodes:
            # print("From node:",str(fromNode.nodeID))
            for toNode in fromNode.connectedTo:
                # print("To node:",str(toNode.nodeID))
                if toNode.parCluster not in self.connectedTo:
                    self.connectedTo.append(toNode.parCluster)
                    if verb==True:
                        print("connecting from cluster: {} to cluster {}".format(str(self.clusterID),str(toNode.parCluster.clusterID)))
                # else:
                    # print("Connection from cluster: {} to cluster {} already exists".format(str(self.clusterID),str(toNode.parCluster.clusterID)))
        return self.connectedTo
    def updateC_DisplayOrder(self,newOrder):
        self.order=newOrder

    def getMaxNodeOrderOfCluster(self):
        max=-1
        for node in self.nodes:
            if(node.order>max):
                max=node.order
        return max




from AhpAnpLib import functions_AHPLib as reqLib
from AhpAnpLib import inputs_AHPLib as cdf_inp
from AhpAnpLib import structs_AHPLib as cdf_str
from AhpAnpLib import calcs_AHPLib as cdf_calc
# import functions_AHPLib as reqLib
# import inputs_AHPLib as cdf_inp
# import structs_AHPLib as cdf_str
# import calcs_AHPLib as cdf_calc
class RatScale:
    id_generator = reqLib.itertools.count(0)
    def __init__(self,scale_name):
        self.name = scale_name
        self.scaleID = next(self.id_generator)
        #members are [name,ideal priority] or [name,1] if no values are given assuming all equal
        self.members=[]
        #val matrix is the pc matrix that would give the ideal priority vector
        self.val_mat=[]

    def defineScaleByValue(self,mat_values=None,verbal=False,*all_values):
        valuesVector=[]
        
        for value in all_values:
            if isinstance(value, list):
                valuesVector.append(value[1])
            else:
                valuesVector.append(1)
        
        valuesIdealVector=cdf_calc.idealVector(valuesVector)
        i=0
        for value in all_values:
            if isinstance(value, list):
                value[1]=valuesIdealVector[i]
                self.members.append(value)
            else:
                self.members.append([value,1])
            i+=1
        if(mat_values is None):
            self.val_mat=cdf_inp.convDirect2PairwiseMatrix(valuesIdealVector)
        else:
            self.val_mat= reqLib.cp.deepcopy(mat_values)

        if(verbal):
            print(all_values)
            print(valuesVector)
            print(self.val_mat)
    
    def __repr__(self):
        labels=[]
        for value in self.members:
            labels.append(value[0])
        scaleDesc="\n-------------------Rating model scale---------\nName:"+ self.name+"\n"+reqLib.tabulate(self.members, headers=['Member name', 'Value'], tablefmt='orgtbl')+"\n--------------------------------------------\n"+reqLib.tabulate(self.val_mat, headers=labels, tablefmt='orgtbl'+"\n")
        return scaleDesc

    def getmIndexBymName(self,mName,verbal=False):
        if verbal: print(f"Checking:{mName}")
        for i,member in enumerate(self.members):
            if verbal: print(f"??={member[0]}")
            if mName==member[0]:
                if verbal:
                    print (f"Found in pos {i}")
                return i
        return -1
    def getmValueBymName(self,mName,verbal=False):
        # print(mName)
        index=self.getmIndexBymName(mName,verbal)
        
        if index!=-1:

            pr_vector=cdf_calc.priorityVector(self.val_mat)
            id_vect=cdf_calc.idealVector(pr_vector)
            if verbal:
                    print (f"Scale: {self.val_mat}\n Priorities {pr_vector} \n Ideal {id_vect}")
            return(id_vect[index])
        else:
            return -1 

class RatStruct:
    def __init__(self,parent):
        self.parent=parent
        #existing model nodes to be used for the ratings model columns (criteria) 
        #each model can have only one rat model attached to it
        self.ratCriteria=[]
        #existing model nodes to be used for the ratings model's rows (alternatives)
        self.ratAlternatives=[]
        #list of ratScales available to this ratings model 
        self.ratScales=[]
        #assignment of scale to criterion
        self.critScaleCombo=[]
        #the ratings table containing the indices of the selected scale items
        self.ratMatrix=[]
        #the priority vector to be used for the calculations - you get it from the limiting priorities re-normalized after removing the unused components
        self.ratCritPriorities=[]

    def __repr__(self):
        return "\n Criteria"+str(self.ratCriteria)+"\n Alternatives to be rated:"+str(self.ratAlternatives)+"\n Scales available to use"+str(self.ratScales)+"\n Criteria and selected scales for each of them:"+str(self.critScaleCombo)
    
    def addCriteriaByVar(self,*args):
        for node in args:
            self.ratCriteria.append(node)
        
    def addCriteriaByName(self,*args):
        for nodeName in args:
            self.ratCriteria.append(self.parent.getNodeObjByName(nodeName))
    
    def addAlternativesByVar(self,*args):
        for node in args:
            self.ratAlternatives.append(node)
        
    def addAlternativesByName(self,*args):
        #if the node does not exist a new node will be created and added to the ratingAlternatives cluster
        for nodeName in args:
            theNode=self.parent.getNodeObjByName(nodeName)
            if(theNode!=-1):
                self.ratAlternatives.append(theNode)
            else:
                #the node does not exist so let's create it
                theAltCluster=self.parent.getClusterIDByName("Alternatives")
                if(theAltCluster==-1):
                    new_nd=cdf_str.Node(nodeName,1+self.parent.getMaxNodeOrder())
                    cl_index=self.parent.getClusterIndexByName("ratingAlternatives")
                    if(cl_index==-1):
                        new_cl=cdf_str.Cluster("ratingAlternatives",1+self.parent.getMaxClusterOrder())
                        new_cl.addNode2Cluster(new_nd)
                        self.parent.addCluster2Model(new_cl)
                    else:
                        self.parent.clusters[cl_index].addNode2Cluster(new_nd)
                self.ratAlternatives.append(new_nd)

    def addScaleByVar(self,*args):
        for scale in args:
            self.ratScales.append(scale)
            # print(self.ratScales)

    def scaleExists(self,name):
        for indx,item in enumerate(self.ratScales):
            if (item.name==name):
                return indx
        return -1
    
    def criterionExists(self,name):
        for indx,item in enumerate(self.ratCriteria):
            if (item.name==name):
                return indx
        return -1
    
    def getScaleObjByName(self,name):

        for scale in self.ratScales:
            if (scale.name==name):
                return scale
        return -1
    def getRatCritObjByName(self,name):
        for item in self.ratCriteria:
            if (item.name==name):
                return item
        return -1
    
    def assignScale2CriterionByName(self,criterionName,scaleName,verbal=False):
        #check if scale exists in the ratings model
        scale_index=self.scaleExists(scaleName)
        #check if criterion belongs to this ratings model
        crit_index=self.criterionExists(criterionName)

        #cannot have 2 criteria connected to the same scale object so if the same create a copy
        found = any(scaleName in tuple_ for tuple_ in self.critScaleCombo)
        #add scale, criterion tuple to the ratings model
        if(found):
            scale2copy=reqLib.cp.deepcopy(self.getScaleObjByName(scaleName))
            new_name=scaleName+"_cp"
            found_nm = any(new_name in scale.name for scale in self.ratScales)
            while(found_nm):
                new_name+="_cp"
            scale2copy.name=new_name
            scale2copy.ID=next(scale2copy.id_generator)
            self.addScaleByVar(scale2copy)
            self.critScaleCombo.append([criterionName,new_name])
        else:
            self.critScaleCombo.append([criterionName,scaleName])
        if (verbal):
            print(f"{scaleName} found in position {scale_index}\n")
            print(f"{criterionName} found in position {crit_index}\n")
            print(f"{self.critScaleCombo} updated\n")

    def getScaleOfCriterion(self, criterion):
        if(self.criterionExists(criterion.name)!=-1):
            
            for pair in self.critScaleCombo:
                if (pair[0]==criterion.name):
                    scale_name=pair[1]
                    return self.getScaleObjByName(scale_name)
        return -1


    def updateScaleMatrix(self,scale_name,new_val_matrix):
        #update the scale matrix and member values with what you just read
        for ind,scale in enumerate(self.ratScales):
            if scale.name==scale_name:
                self.ratScales[ind].val_mat=new_val_matrix
                pr_vector=cdf_calc.priorityVector(new_val_matrix)
                id_vect=cdf_calc.idealVector(pr_vector)
                for i,member in enumerate(scale.members):
                    member[1]=id_vect[i]
                # print(scale.members)


from AhpAnpLib import functions_AHPLib as reqLib
from AhpAnpLib import calcs_AHPLib as cdf_calc
from AhpAnpLib import ratings_AHPLib as cdf_rat
from AhpAnpLib import structs_AHPLib as cdf_str
# import functions_AHPLib as reqLib
# import calcs_AHPLib as cdf_calc
# import ratings_AHPLib as cdf_rat
# import structs_AHPLib as cdf_str
def genFullQuest(model,keyword,verb=False):
    questnr=[]
    for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 

        for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
            
            connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
            for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                i=0
                qset=[]
                # print("Nodes:\n",clusterPWC.nodes)
                for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                    j=0
                    # print("NodeA: ",nodeA.nodeID)
                    for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        # print("NodeB: ",nodeB.nodeID)
                        if (nodeA.nodeID!= nodeB.nodeID) and (i<j) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                            # print("{}-{}".format(i,j))
                            quest="With respect to {}, which one is more {}: {} or {} ? By how much?".format(nodeFrom.name, keyword, nodeA.name, nodeB.name)
                            if verb:
                                print(quest)
                            qset.append(quest)
                        j+=1
                    i+=1
                questnr.append(qset)
                if verb:
                    print("---------------------------------------------------------------------------\n")
    return questnr
def genFirstLineAboveDiagQuest(model,keyword,verb=False):
    questnr=[]
    for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 

        for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
            
            connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
            for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                i=0
                qset=[]
                # print("Nodes:\n",clusterPWC.nodes)
                for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                    j=0
                    # print("NodeA: ",nodeA.name)
                    for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        # print("NodeB: ",nodeB.name)
                        # print("{}-{}".format(i,j),nodeA in nodeFrom.connectedTo,nodeB in nodeFrom.connectedTo)
                        if (nodeA.nodeID!= nodeB.nodeID) and ((i==0)or (i+1==j)) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                            
                            quest="With respect to {}, which one is more {}: {} or {} ? By how much?".format(nodeFrom.name, keyword, nodeA.name, nodeB.name)
                            if verb:
                                print(quest)
                            qset.append(quest)
                        j+=1
                    i+=1
                questnr.append(qset)
                if verb:
                    print("---------------------------------------------------------------------------\n")
    return questnr
def genFirstLineQuest(model,keyword,verb=False):
    questnr=[]
    for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 

        for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
            
            connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
            for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                i=0
                qset=[]
                # print("Nodes:\n",clusterPWC.nodes)
                nodeA=clusterPWC.nodes[0]
                j=0
                # print("NodeA: ",nodeA.name)
                for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                    # print("NodeB: ",nodeB.name)
                    # print("{}-{}".format(i,j),nodeA in nodeFrom.connectedTo,nodeB in nodeFrom.connectedTo)
                    if (nodeA.nodeID!= nodeB.nodeID) and (i<j) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                        # print("{}-{}".format(i,j))
                        quest="With respect to {}, which one is more {}: {} or {} ? By how much?".format(nodeFrom.name, keyword, nodeA.name, nodeB.name)
                        if verb:
                            print(quest)
                        qset.append(quest)
                    j+=1
                questnr.append(qset)
                if verb:
                 print("---------------------------------------------------------------------------\n")
                
    return questnr

def genexport4QualtricsQuestFull(filepath,model,keyword,howMuch=True):
    count=1
    with open(filepath, 'w', encoding='UTF8',newline='') as f: 
        f.write("[[AdvancedFormat]]\n")
        for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 

            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
                f.write("[[Block: With respect to: "+nodeFrom.name+"]]\n")
                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    i=0

                    # print("Nodes:\n",clusterPWC.nodes)
                    for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        j=0
                        # print("NodeA: ",nodeA.nodeID)
                        for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                            # print("NodeB: ",nodeB.nodeID)
                            if (nodeA.nodeID!= nodeB.nodeID) and (i<j) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                                # print("{}-{}".format(i,j))
                            
                                f.write("\n[[Question:MC:SingleAnswer]]\n")
                                quest1_csv=str(count)+'. With respect to '+nodeFrom.name+ ' which one is more '+keyword+':'+'\n'
                                f.write(quest1_csv) 
                                f.write("\n[[Choices]]\n")
                                
                                ans1=nodeA.name+'\n'
                                f.write(ans1)
                                ans2=nodeB.name+'\n'
                                f.write(ans2)
                                count+=1
                                
                                if(howMuch):
                                    f.write("\n[[Question:MC:Dropdown]]\n")
                                    quest2_csv=str(count)+". By how much?"+'\n'
                                    f.write(quest2_csv)
                                    f.write("\n[[Choices]]\n")
                                    f.write("1\n2\n3\n4\n5\n6\n7\n8\n9\n")
                                    count+=1
                                f.write('\n')
                                
                            j+=1
                        i+=1
def genexport4QualtricsFirstLineAboveDiagQuest(filepath,model,keyword,howMuch=True):
    count=1
    with open(filepath, 'w', encoding='UTF8',newline='') as f:
        f.write("[[AdvancedFormat]]\n")

        for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 

            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
                f.write("[[Block: With respect to: "+nodeFrom.name+"]]\n")
                
                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    i=0
                    # f.write("[[Block:"+clusterPWC.name+"]]\n")
                    # print("Nodes:\n",clusterPWC.nodes)
                    for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        j=0
                        # print("NodeA: ",nodeA.nodeID)
                        for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                            # print("NodeB: ",nodeB.nodeID)
                            if (nodeA.nodeID!= nodeB.nodeID) and ((i==0)or (i+1==j)) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                                f.write("\n[[Question:MC:SingleAnswer]]\n")
                                quest1_csv=str(count)+'. With respect to '+nodeFrom.name+ ' which one is more '+keyword+':'+'\n'
                                f.write(quest1_csv) 
                                f.write("\n[[Choices]]\n")
                                
                                ans1=nodeA.name+'\n'
                                f.write(ans1)
                                ans2=nodeB.name+'\n'
                                f.write(ans2)
                                count+=1
                                
                                if(howMuch):
                                    f.write("\n[[Question:MC:Dropdown]]\n")
                                    quest2_csv=str(count)+". By how much?"+'\n'
                                    f.write(quest2_csv)
                                    f.write("\n[[Choices]]\n")
                                    f.write("1\n2\n3\n4\n5\n6\n7\n8\n9\n")
                                    count+=1
                                f.write('\n')
                            j+=1
                        i+=1       
def genexport4QualtricsFirstLineQuest(filepath,model,keyword,howMuch=True):
    count=1
    with open(filepath, 'w', encoding='UTF8',newline='') as f:
        f.write("[[AdvancedFormat]]\n")
        for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 
            
            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
                f.write("[[Block: With respect to: "+nodeFrom.name+"]]\n")
                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    i=0
                    qset=[]
                    # print("Nodes:\n",clusterPWC.nodes)
                    nodeA=clusterPWC.nodes[0]
                    j=0
                    # print("NodeA: ",nodeA.nodeID)
                    for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        # print("NodeB: ",nodeB.nodeID)
                        if (nodeA.nodeID!= nodeB.nodeID) and (i<j) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                                f.write("\n[[Question:MC:SingleAnswer]]\n")
                                quest1_csv=str(count)+'. With respect to '+nodeFrom.name+ ' which one is more '+keyword+':'+'\n'
                                f.write(quest1_csv) 
                                f.write("\n[[Choices]]\n")
                                
                                ans1=nodeA.name+'\n'
                                f.write(ans1)
                                ans2=nodeB.name+'\n'
                                f.write(ans2)
                                count+=1
                                
                                if(howMuch):
                                    f.write("\n[[Question:MC:Dropdown]]\n")
                                    quest2_csv=str(count)+". By how much?"+'\n'
                                    f.write(quest2_csv)
                                    f.write("\n[[Choices]]\n")
                                    f.write("1\n2\n3\n4\n5\n6\n7\n8\n9\n")
                                    count+=1
                                f.write('\n')
                        j+=1

def genexport4GoogleQuestFull(filepath,model,keyword,howMuch=True):
    count=1
    with open(filepath, 'w', encoding='UTF8',newline='') as f:
        writer = reqLib.csv.writer(f)  
        for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 

            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
                
                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    i=0

                    # print("Nodes:\n",clusterPWC.nodes)
                    for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        j=0
                        # print("NodeA: ",nodeA.nodeID)
                        for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                            # print("NodeB: ",nodeB.nodeID)
                            if (nodeA.nodeID!= nodeB.nodeID) and (i<j) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                                # print("{}-{}".format(i,j))
                                quest1_csv=["With respect to "+nodeFrom.name+' which one is more '+keyword, nodeA.name, nodeB.name]
                                writer.writerow(quest1_csv)  
                                quest2_csv=["By how much?","1","2","3","4","5","6","7","8","9"]
                                writer.writerow(quest2_csv)                            
                            j+=1
                        i+=1
def genexport4GoogleFirstLineAboveDiagQuest(filepath,model,keyword,howMuch=True):
    count=1
    with open(filepath, 'w', encoding='UTF8',newline='') as f:
        writer = reqLib.csv.writer(f)  

        for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 

            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
                
                
                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    i=0
                    # f.write("[[Block:"+clusterPWC.name+"]]\n")
                    # print("Nodes:\n",clusterPWC.nodes)
                    for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        j=0
                        # print("NodeA: ",nodeA.nodeID)
                        for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                            # print("NodeB: ",nodeB.nodeID)
                            if (nodeA.nodeID!= nodeB.nodeID) and ((i==0)or (i+1==j)) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                                quest1_csv=["With respect to "+nodeFrom.name+' which one is more '+keyword, nodeA.name, nodeB.name]
                                writer.writerow(quest1_csv)  
                                quest2_csv=["By how much?","1","2","3","4","5","6","7","8","9"]
                                writer.writerow(quest2_csv)  
                            j+=1
                        i+=1       
def genexport4GoogleFirstLineQuest(filepath,model,keyword,howMuch=True):
    count=1
    with open(filepath, 'w', encoding='UTF8',newline='') as f:
        writer = reqLib.csv.writer(f)  
        for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')): 
            
            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
               
                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    i=0
                    qset=[]
                    # print("Nodes:\n",clusterPWC.nodes)
                    nodeA=clusterPWC.nodes[0]
                    j=0
                    # print("NodeA: ",nodeA.nodeID)
                    for nodeB in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        # print("NodeB: ",nodeB.nodeID)
                        if (nodeA.nodeID!= nodeB.nodeID) and (i<j) and (nodeA in nodeFrom.connectedTo) and (nodeB in nodeFrom.connectedTo):
                                quest1_csv=["With respect to "+nodeFrom.name+' which one is more '+keyword, nodeA.name, nodeB.name]
                                writer.writerow(quest1_csv)  
                                quest2_csv=["By how much?","1","2","3","4","5","6","7","8","9"]
                                writer.writerow(quest2_csv)  
                        j+=1

def export4ExcelQuestFull(model,filepath,show_estimates=False,verb=False):
    try:
        workbook = reqLib.xlsxwriter.Workbook(filepath)
        worksheet = workbook.add_worksheet("pairwise_comp")
        worksheet.set_column(0,20,15)

        cell_format_hdTitle  = workbook.add_format({'bold': True,'font_color': 'red', 'font_size':16})
        cell_format_hdTitle.set_pattern(1)
        cell_format_hdTitle.set_bg_color('C3D5FF')
        cell_format_hd  = workbook.add_format({'bold': True,'font_color': 'red'})
        cell_format_hd.set_pattern(1)
        cell_format_hd.set_bg_color('C3D5FF')
        cell_format_hd2  = workbook.add_format({'bold': True,'font_color': 'red'})
        cell_format_hd2.set_pattern(1)
        cell_format_hd2.set_bg_color('#d5ffc3')
        cell_format_hd3  = workbook.add_format({'bold': True,'font_color': 'red'})
        
        
        cell_format_ln  = workbook.add_format({'bold': True,'font_color': 'blue'})
        cell_format_ln.set_border(1)
        cell_format_diag=workbook.add_format()
        cell_format_diag.set_pattern(1)
        cell_format_diag.set_border(1)
        cell_format_diag.set_bg_color('FFFF6B')
        cell_format_block=workbook.add_format()
        cell_format_block.set_pattern(1)
        cell_format_block.set_border(1)
        cell_format_block.set_bg_color('#808080')
        cell_format_dir=workbook.add_format()
        cell_format_dir.set_pattern(1)
        cell_format_dir.set_border(1)
        cell_format_dir.set_bg_color('#ceebd6')
        cell_format_empty  = workbook.add_format()
        cell_format_empty.set_border(1)
        
        decimal_format = workbook.add_format({'num_format': '0.00','bg_color': '#accbe8',  'border': '1'})
        res_hd_format = workbook.add_format({'bold': True,'font_color': 'blue','bg_color': '#accbe8',  'border': '1'})
        inc_hd_format=workbook.add_format({'bold': True,'font_color': 'orange','bg_color': '#f9d5b6',  'border': '1'})
        inc_decimal_format = workbook.add_format({'num_format': '0.00','bg_color': '#f9d5b6',  'border': '1'})
        row=0
        col=0
        for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')):
            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 

                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                if(len(connectedClusters)>0):
                    worksheet.write(row, 0, nodeFrom.name,cell_format_hdTitle)
                    if(verb):
                        print(row,nodeFrom.name)
                    row=row+1
                    for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                        if(verb):
                            print(row,"Enter judgments:\n",row+1,clusterPWC.name)
                        worksheet.write(row, 0, "Enter judgments for the paiwise comparisons in the matrix or direct values in the green cells",cell_format_hd3)
                        worksheet.write(row+1, 0, clusterPWC.name,cell_format_hd2)
                        row=row+1
                        col=0
                        size=0
                        if(verb):
                            print(row,"Now nodes\n")
                        for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                            # print("NodeA: ",nodeA.nodeID)         
                            
                            if(nodeA in nodeFrom.connectedTo):
                                worksheet.write(row, col+1, nodeA.name,cell_format_ln)
                                worksheet.write(row+col+1, 0, nodeA.name,cell_format_ln)
                                worksheet.write_number(row+col+1, col+1,1.000 ,cell_format_diag)
                                # print(nodeA.name)
                                col+=1
                                size+=1
                            worksheet.write(row, size+1, "Direct values",cell_format_ln)
                            # if(verb):
                            #     print(row,"Direct")
                        for a in range(size):
                            for b in range (size):
                                if(a!=b and a>b):
                                    val='=1/'+convIndex2ExcelRef(row+b+2,a+2)
                                    worksheet.write(row+a+1,b+1,val,cell_format_block)
                                if(a!=b and a<b):
                                    worksheet.write(row+a+1,b+1,"",cell_format_empty)
                            worksheet.write(row+a+1,size+1,"",cell_format_dir)
                        if(show_estimates==True):
                            worksheet.write(row,col+2,"Line Sum",res_hd_format)
                            worksheet.write(row,col+3,"Estimated Priority",res_hd_format)
                            worksheet.write(row+size+1,0,"Sum of Col",inc_hd_format)
                            worksheet.write(row+size+2,col+2,"Est. Incons.",inc_hd_format)
                            for a in range(size):
                                val='='
                                valColSum=''
                                for b in range (size):
                                    val+='+'+convIndex2ExcelRef(row+a+2,b+2)
                                    valColSum='=sum('+convIndex2ExcelRef(row+2,b+2)+':'+convIndex2ExcelRef(row+2+size-1,b+2)+')'
                                    worksheet.write(row+size+1,b+1,valColSum,inc_decimal_format)
                                worksheet.write(row+a+1,col+2,val,decimal_format)
                                
                            valRowSum='=sum('+convIndex2ExcelRef(row+2,col+3)+':'+convIndex2ExcelRef(row+2+size-1,col+3)+')'

                            worksheet.write(row+a+2,col+2,valRowSum,decimal_format)
                            
                            #calc tem priorities
                            for a in range(size):
                                val='='+convIndex2ExcelRef(row+a+2,col+3)+'/'+convIndex2ExcelRef(row+size+2,col+3)
                                worksheet.write(row+a+1,col+3,val,decimal_format)
                            #calc inconsistency
                            #ci=(lmax-size)/(size-1)
                            priorRef=convIndex2ExcelRef(row+2,col+4)+':'+convIndex2ExcelRef(row+2+size-1,col+4)
                            colSumRef=convIndex2ExcelRef(row+size+2,2)+':'+convIndex2ExcelRef(row+size+2,size+1)
                            valInc='=((MMULT('+colSumRef+','+priorRef+')-'+str(size)+')/('+str(size)+'-1))/'+str(cdf_calc.RI(size))
                            worksheet.write(row+size+2,col+3,valInc,inc_decimal_format)
                        # if verb:
                        #     print(nodeA.name,size)        
                        
                        row=row+col+3
                        if verb:
                            print("Saved pairwise comparison matrix for cluster : ",clusterPWC.name, "with respect to node:", nodeFrom.name)
                            print(f"Next row: {row}")    
                    # row+=1
                    if verb:
                            print(f"Row {row} {cluster.name} {nodeFrom.name}")   
        workbook.close()
    except Exception as e:
        print(f"An error occured while trying to save the pairwise comparison matrices to the file. Check if the file is already open in Excel and if yes, then please close it and retry.\n Error: {e}")


def export4ExcelRatingScales(model,filepath,load_values=True, show_estimates=False, verb=False):
    # print(verb)
    if(verb):
        print("setting up rating scales comparison matrices ")
    border = reqLib.Border(left=reqLib.Side(style='thin'), right=reqLib.Side(style='thin'), top=reqLib.Side(style='thin'), bottom=reqLib.Side(style='thin'))
    # border_noright = reqLib.Border(left=reqLib.Side(style='thin'), right=reqLib.Side(style=None), top=reqLib.Side(style='thin'), bottom=reqLib.Side(style='thin'))
    # border_noleft = reqLib.Border(left=reqLib.Side(style=None), right=reqLib.Side(style='thin'), top=reqLib.Side(style='thin'), bottom=reqLib.Side(style='thin'))
    lower_diag_fill = reqLib.PatternFill(start_color="bcbcbc", end_color="bcbcbc", fill_type="solid")
    res_hd_fill = reqLib.PatternFill(start_color="ffb6e5f2", end_color="ffb6e5f2", fill_type="solid")
    dir_hd_fill = reqLib.PatternFill(start_color="6aa84f", end_color="6aa84f", fill_type="solid")
    dir_fill = reqLib.PatternFill(start_color="b6d7a8", end_color="b6d7a8", fill_type="solid")
    
    diag_light = reqLib.PatternFill(start_color="FFFF6B", end_color="FFFF6B", fill_type="solid")
    incons = reqLib.PatternFill(start_color="f9d5b6", end_color="f9d5b6", fill_type="solid")
    estimates= reqLib.PatternFill(start_color="accbe8", end_color="accbe8", fill_type="solid")
    bold_font=reqLib.Font(bold=True)
    # rule_n = reqLib.DataBarRule(start_type='min', start_value=0, end_type='max', end_value=0,
    #                color=reqLib.Color(rgb='fff2463a'), showValue=True)
    # rule_i = reqLib.DataBarRule(start_type='min', start_value=0, end_type='max', end_value=0,
    #                color=reqLib.Color(rgb='ff0b30b5'), showValue=True)
    
    worksheetName="rating_scales"
    try:
        if reqLib.os.path.isfile(filepath):
            #print("load")
            workbook = reqLib.openpyxl.load_workbook(filepath)
        else:
            #print("create")
            workbook = reqLib.openpyxl.Workbook()
            if 'Sheet' in workbook.sheetnames:
                del workbook['Sheet']
            if worksheetName not in workbook.sheetnames:
                workbook.create_sheet(title=worksheetName)

        workbook.active = workbook.sheetnames.index(worksheetName)
        worksheet = workbook.active
          
        row=1
        col=1
        if(model.modelType=="ratings"):
            for my_crit in model.rateModel.ratCriteria:
                #the criterion name as a title
                worksheet.cell(row, col).value= my_crit.name
                worksheet.cell(row, col).fill = res_hd_fill
                worksheet.cell(row, col).font = bold_font
                hd_row=row
                row+=1
                my_scale=model.rateModel.getScaleOfCriterion(my_crit)
                if(verb):
                    print(f"Criterion: {my_crit}")
                    print(my_scale)
                print("Error!!!?",my_scale,my_crit)
                print(my_scale.members)
                size=0
                #the scale matrix
                for item in my_scale.members:
                    worksheet.cell(hd_row, col+1).fill = res_hd_fill

                    worksheet.cell(row, col+1).value= item[0]
                    worksheet.cell(row, col+1).border = border
                    worksheet.cell(row, col+1).font = bold_font
                    
                    worksheet.cell(row+col,1).value= item[0]
                    worksheet.cell(row+col,1).border = border
                    worksheet.cell(row+col,1).font = bold_font
                    col+=1
                    size+=1
                
                # direct values
                worksheet.cell(hd_row, col+1).fill = res_hd_fill
                worksheet.cell(row, col+1).fill =dir_hd_fill
                worksheet.cell(row, col+1).value ="Direct values"
                worksheet.cell(row, col+1).border = border
                worksheet.cell(row, col+1).font = bold_font
             
                #inside the table
                for inside_i in range(size+1):
                    worksheet.cell(row+inside_i, col+1).fill = dir_fill
                    worksheet.cell(row+inside_i, col+1).border = border
                    for inside_j in range(size+1):
                        worksheet.cell(row+inside_i, inside_j+1).border = border
                        if(load_values==True & (inside_i>0)&(inside_j>0)&(inside_i<inside_j)):
                            worksheet.cell(row+inside_i, inside_j+1).value=my_scale.val_mat[inside_i-1,inside_j-1]
                            worksheet.cell(row+inside_i, inside_j+1).number_format = '0.000'
                        if((inside_i==inside_j) &(inside_i>0)):
                            worksheet.cell(row+inside_i, inside_j+1).fill=diag_light
                            worksheet.cell(row+inside_i, inside_j+1).value=1.0
                        if((inside_i>inside_j) &(inside_i>0)):
                            worksheet.cell(row+inside_i, inside_j+2).fill=lower_diag_fill
                            # worksheet.cell(row+inside_i, inside_j+2).protection = reqLib.Protection(locked=True)
                            worksheet.cell(row+inside_i, inside_j+2).value='=1/'+convIndex2ExcelRef(row+inside_j+1,inside_i+1)
                            worksheet.cell(row+inside_i, inside_j+2).number_format = '0.000'
                #show estimates
                if(show_estimates==True):
                    set_cell_value_and_style(worksheet.cell(row,col+2),"Line Sum",estimates,border, "ff123ef1" ,True)
                    set_cell_value_and_style(worksheet.cell(row,col+3),"Est. Normal Priorities",estimates,border,"ff123ef1",True)
                    set_cell_value_and_style(worksheet.cell(row,col+4),"Est. Ideal Priorities",estimates,border,"ff123ef1",True)
                    set_cell_value_and_style(worksheet.cell(row+size+1,1),"Sum of Col",incons,border,"ff000000",True)
                    set_cell_value_and_style(worksheet.cell(row+size+2,col+2),"Est. Incons.",incons,border,"ff000000",True)

                    for a in range(size):
                        val='='
                        valColSum=''
                        for b in range (size):
                            val+='+'+convIndex2ExcelRef(row+a+1,b+2)
                            valColSum='=sum('+convIndex2ExcelRef(row+1,b+2)+':'+convIndex2ExcelRef(row+size,b+2)+')'
                            set_cell_value_and_style(worksheet.cell(row+size+1,b+2),valColSum,incons,border, "ff000000" ,True)
                            worksheet.cell(row+size+1,b+2).number_format = '0.000'
                        set_cell_value_and_style(worksheet.cell(row+a+1,col+2),val,estimates,border, "ff123ef1" ,False)
                        worksheet.cell(row+a+1,col+2).number_format = '0.000'
                    valRowSum='=sum('+convIndex2ExcelRef(row+1,col+2)+':'+convIndex2ExcelRef(row+size,col+2)+')'
                    set_cell_value_and_style(worksheet.cell(row+a+2,col+2),valRowSum,estimates,border, "ff123ef1" ,False)
                    worksheet.cell(row+a+2,col+2).number_format = '0.000'
                    #calc tem priorities
                    for a in range(size):
                        val='='+convIndex2ExcelRef(row+a+1,col+2)+'/'+convIndex2ExcelRef(row+size+1,col+2)
                        ideal_val='='+convIndex2ExcelRef(row+a+1,col+3)+'/MAX('+convIndex2ExcelRef(row+1,col+3)+':'+convIndex2ExcelRef(row+size+1,col+3)+')'
                        set_cell_value_and_style(worksheet.cell(row+a+1,col+3),val,estimates,border, "ff123ef1" ,True)
                        set_cell_value_and_style(worksheet.cell(row+a+1,col+4),ideal_val,estimates,border, "ff123ef1" ,True)
                        worksheet.cell(row+a+1,col+3).number_format = '0.000'
                        worksheet.cell(row+a+1,col+4).number_format = '0.000'
                    #calc inconsistency
                    #ci=(lmax-size)/(size-1)
                    priorRef=convIndex2ExcelRef(row+1,col+3)+':'+convIndex2ExcelRef(row+size,col+3)
                    colSumRef=convIndex2ExcelRef(row+size+1,2)+':'+convIndex2ExcelRef(row+size+1,size+1)
                    valInc='=((MMULT('+colSumRef+','+priorRef+')-'+str(size)+')/('+str(size)+'-1))/'+str(cdf_calc.RI(size))
                    set_cell_value_and_style(worksheet.cell(row+size+2,col+3),valInc,incons,border, "ff000000" ,True)
                    worksheet.cell(row+size+2,col+3).number_format = '0.000'
                #end of show estimates
                row=row+size+4
                col=1
            for column in worksheet.columns:
                column_letter = column[0].column_letter
                worksheet.column_dimensions[column_letter].width = 20
            workbook.save(filepath)
 
    except Exception as e:
        print(f"An error occured while trying to save the rating scale tables to the file. Check if the file is already open in Excel and if yes, then please close it and retry.\n Error: {e}")


def export4ExcelRatingsTable(model,filepath,verb=False):

    worksheetName="rating_table"
    
    bold_font=reqLib.Font(bold=True)
    res_hd_fill = reqLib.PatternFill(start_color="ffb6e5f2", end_color="ffb6e5f2", fill_type="solid")
    res_cr_fill = reqLib.PatternFill(start_color="ffcde4f7",fill_type="solid")
    tot_res_fill = reqLib.PatternFill(start_color="ffffffcc",fill_type="solid")
    tot_hd_fill = reqLib.PatternFill(start_color="fff2fa07",fill_type="solid")
    pr_res_fill = reqLib.PatternFill(start_color="ffffffb3",fill_type="solid")
    border = reqLib.Border(left=reqLib.Side(style='thin'), right=reqLib.Side(style='thin'), top=reqLib.Side(style='thin'), bottom=reqLib.Side(style='thin'))
    no_border = reqLib.Border()
    incons = reqLib.PatternFill(start_color="f9d5b6", end_color="f9d5b6", fill_type="solid")
    results_fill = reqLib.PatternFill(start_color="b6d7a8", end_color="b6d7a8", fill_type="solid")
    try:
         if reqLib.os.path.isfile(filepath):
             print("load")
             workbook = reqLib.openpyxl.load_workbook(filepath)
         else:
             print("create")
             workbook = reqLib.openpyxl.Workbook()
            #  if 'Sheet' in workbook.sheetnames:
            #      del workbook['Sheet']
         if worksheetName not in workbook.sheetnames:
            workbook.create_sheet(title=worksheetName)

         workbook.active = workbook.sheetnames.index(worksheetName)
         worksheet = workbook.active
        
         row=1
         col=1
         if(model.modelType=="ratings"):
             #rows: alternatives
             #cols: sel. criteria
             #cells: scale of selected criterion
             prev_end=-3
             for ind,my_crit in enumerate(model.rateModel.ratCriteria):
                 # print(ind,my_crit)
                 #this is empty
                 # print(model.rateModel.ratCritPriorities)
                 len_rat_tbl=3+len(model.rateModel.ratAlternatives)
                 col+=1
                 #the criterion name as a title
                 set_cell_value_and_style(worksheet.cell(row,col), my_crit.name,res_hd_fill,border, "ff000000" ,True)
                 worksheet.cell(row+len_rat_tbl+2,col).value=model.rateModel.ratCritPriorities[ind]
                 #"=INDEX(rating_scales!"+convIndex2ExcelRef(ref_from,col_res)+":"+convIndex2ExcelRef(ref_to,col_res)+", MATCH("+convIndex2ExcelRef(in_row,col)+",rating_scales!"+c_from+":"+c_to+", 0))"
                 worksheet.cell(row+len_rat_tbl+2,col).border = border
                 worksheet.cell(row+len_rat_tbl+2,col).number_format = '0.000'
                
                 #add scale cells
                 my_scale=model.rateModel.getScaleOfCriterion(my_crit)
                 ref_from=6+prev_end
                 c_from="A"+str(ref_from)
                 ref_to=6+prev_end+len(my_scale.members)-1
                 c_to="A"+str(ref_to)
                 
                 prev_end=6+prev_end+len(my_scale.members)-1
                 # print(f"from:{c_from}-ti:{c_to}-prev:{prev_end}")
                 in_row=2
                 
                 col_res=len(my_scale.members)+5
                 for my_alt in model.rateModel.ratAlternatives:
                     worksheet=addDropDown2ExcelCell(worksheet,c_from,c_to,"rating_scales",convIndex2ExcelRef(in_row,col))
                     worksheet.cell(in_row,col).fill = incons
                     worksheet.cell(in_row,col).border = border

                     # excel did not recognize xlookup and vlookup worksheet.cell(in_row+len_rat_tbl+2,col).value= "=VLOOKUP("+convIndex2ExcelRef(in_row,col)+",rating_scales!"+c_from+":"+convIndex2ExcelRef(ref_to,col_res)+","+str(col_res)+")"
                     worksheet.cell(in_row+len_rat_tbl+3,col).value= "=INDEX(rating_scales!"+convIndex2ExcelRef(ref_from,col_res)+":"+convIndex2ExcelRef(ref_to,col_res)+", MATCH("+convIndex2ExcelRef(in_row,col)+",rating_scales!"+c_from+":"+c_to+", 0))"
                     worksheet.cell(in_row+len_rat_tbl+3,col).border = border
                     worksheet.cell(in_row+len_rat_tbl+3,col).number_format = '0.000'
                     in_row+=1
                 set_cell_value_and_style(worksheet.cell(in_row+2,col), "",results_fill,no_border, "ff000000" ,True)
                 #the criteria priorities
                #  set_cell_value_and_style(worksheet.cell(in_row+2,col), "value",res_cr_fill,border, "ff000000" ,True)
                 #the criterion name as a title
                 set_cell_value_and_style(worksheet.cell(in_row+5,col), my_crit.name,res_hd_fill,border, "ff000000" ,True)
             col=1
             row+=1   
             num_crit=len(model.rateModel.ratCriteria)
             num_alts=len(model.rateModel.ratAlternatives)
             row_start=row
             for my_alt in model.rateModel.ratAlternatives:
                 #row heading tbl 1
                 set_cell_value_and_style(worksheet.cell(row,col), my_alt.name,res_hd_fill,border, "ff000000" ,True)
                 #row heading tbl 2
                 set_cell_value_and_style(worksheet.cell(row+len_rat_tbl+3,col), my_alt.name,res_hd_fill,border, "ff000000" ,True)
                 #results totals (estimates)
                 valRowSum='=sumproduct('+convIndex2ExcelRef(row+len_rat_tbl+3,col+1)+':'+convIndex2ExcelRef(row+len_rat_tbl+3,col+num_crit)+','+convIndex2ExcelRef(row_start+len_rat_tbl+1,col+1)+':'+convIndex2ExcelRef(row_start+len_rat_tbl+1,col+num_crit)+')'
                 set_cell_value_and_style(worksheet.cell(row+len_rat_tbl+3,col+num_crit+1),valRowSum,tot_res_fill,border, "ff123ef1" ,False)
                 worksheet.cell(row+len_rat_tbl+3,col+num_crit+1).number_format = '0.000'
                
                 #results totals (priorities)
                 
                 valRowPr='='+convIndex2ExcelRef(row+len_rat_tbl+3,col+num_crit+1)+'/sum('+convIndex2ExcelRef(row_start+len_rat_tbl+3,col+num_crit+1)+':'+convIndex2ExcelRef(row_start+len_rat_tbl+2+num_alts,col+num_crit+1)+')'
                 set_cell_value_and_style(worksheet.cell(row+len_rat_tbl+3,col+num_crit+2),valRowPr,pr_res_fill,border, "ff000000" ,False)
                 worksheet.cell(row+len_rat_tbl+3,col+num_crit+2).number_format = '0.000'
                 row+=1
             #write results table
             # =XLOOKUP(B2,rating_scales!$A$3:$A$6,rating_scales!$H$3:$H$6)    
             col=1
             row+=2
             set_cell_value_and_style(worksheet.cell(row,col), "ESTIMATED RESULTS",results_fill,no_border, "ff000000" ,True)
             set_cell_value_and_style(worksheet.cell(row+2,col+num_crit+1), "TOTALS",tot_hd_fill,border, "ff000000" ,True)
             set_cell_value_and_style(worksheet.cell(row+2,col+num_crit+2), "PRIORITIES",tot_hd_fill,border, "ff000000" ,True)
             set_cell_value_and_style(worksheet.cell(in_row+2,col+num_crit+1), "",results_fill,no_border, "ff000000" ,True)
             set_cell_value_and_style(worksheet.cell(in_row+2,col+num_crit+2), "",results_fill,no_border, "ff000000" ,True)
             row+=1
            
         for column in worksheet.columns:
             column_letter = column[0].column_letter
             worksheet.column_dimensions[column_letter].width = 20
         workbook.save(filepath)
            
    
    
    except Exception as e:
        print(f"An error occured while trying to save the rating table to the file. Check if the file is already open in Excel and if yes, then please close it and retry.\n Error: {e}")

def export4ExcelRatingsSetup(model,exportfile,show_estimates=False,verb=False):
    cdf_calc.calcRateCritV(model,verbal=verb)
    export4ExcelRatingScales(model,exportfile,True,show_estimates,False)
    export4ExcelRatingsTable(model,exportfile,False)
    cdf_calc.updateRateTblResults2Excel(model,exportfile,"rating_table",verb=verb)

def calcExcelRatings(model,importfile,exportfile,verb=False):
    importRatingsFromExcel(model,importfile,"rating_scales",verb)    
    importRatTableFromExcel(model,importfile,"rating_table",verb)  
    calcRatResults(model,exportfile,verb)

def addLabels2Excel(label_list,filepath,wsheet_name,start_row,start_col,direction):
    #direction 0 horizontal, 1 vertical, 2 both
    bold_font=reqLib.Font(bold=True)
    res_hd_fill = reqLib.PatternFill(start_color="ffb6e5f2", end_color="ffb6e5f2", fill_type="solid")
    border = reqLib.Border(left=reqLib.Side(style='thin'), right=reqLib.Side(style='thin'), top=reqLib.Side(style='thin'), bottom=reqLib.Side(style='thin'))
    no_border = reqLib.Border()
    
    try:
        if reqLib.os.path.isfile(filepath):
            # print("load")
            workbook = reqLib.openpyxl.load_workbook(filepath)
            worksheet = workbook[wsheet_name]
            
            #worksheet = workbook.active
            row=start_row
            col=start_col
            for i,label in enumerate(label_list):
                if(direction==0):
                    if(label!=label_list[i-1]):
                        set_cell_value_and_style(worksheet.cell(row+i,col), label,res_hd_fill,border, "ff000000" ,True)
                    else:
                        set_cell_value_and_style(worksheet.cell(row+i,col), "",res_hd_fill,no_border, "ff000000" ,True)
                if(direction==1):
                    if(label!=label_list[i-1]):
                        set_cell_value_and_style(worksheet.cell(row,col+i), label,res_hd_fill,border, "ff000000" ,True)
                    else:
                        set_cell_value_and_style(worksheet.cell(row,col+i), "",res_hd_fill,no_border, "ff000000" ,True)
                # print(f"label{label},row{row+i}, col{col+i}\n")
                       
            for column in worksheet.columns:
                column_letter = column[0].column_letter
                worksheet.column_dimensions[column_letter].width = 15
            max_col_dir=3
            if (direction==0):
                max_col_dir=len(label_list)+2
            for row in worksheet.iter_rows(min_row=1, min_col=2, max_row=len(label_list)+2, max_col=max_col_dir):
                for cell in row:
                    if cell.value is not None:
                        cell.border = border
            workbook.save(filepath)
    except Exception as e:
        print(f"An error occured while trying to save the cluster labelsto the file. Check if the file is already open in Excel and if yes, then please close it and retry.\n Error: {e}")

def importFromExcel(model,filepath,sheet_name, verbal=False):

    all_excel = reqLib.pd.read_excel (filepath, sheet_name=sheet_name, header=None)
    array_excel=all_excel.fillna(0)
    file_len=len(array_excel)
    if(verbal==True):
        print("File Len=", file_len)
        print("excel",array_excel)
    row=0
    for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')):
        if(verbal==True): print("Cluster:",cluster.name)
        for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
            connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)

            if(len(connectedClusters)>0):
                if(verbal==True): print("NFrom:",nodeFrom.name)
                if(verbal==True): print("Connected clusters",connectedClusters)
                model.wrtlabels.append(array_excel.iat[row,0])
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    row+=2
                    if(verbal==True): print("From file row: ",row,"working in cluster:",array_excel.iat[row,0],clusterPWC.name)
                    model.clabels.append(clusterPWC.name) #not in excel anymore
                    row=row+1
                    size=0
                    for nodeA in sorted(clusterPWC.nodes, key=reqLib.op.attrgetter('order')):
                        if(nodeA in nodeFrom.connectedTo):
                            model.nlabels.append(array_excel.iat[row+size,0])
                            size+=1

                        if(verbal==True): print("Current row",row,"All node labels till this point:",model.nlabels)
                    pc_matrix =  reqLib.np.empty([size, size], dtype = float)
                    #check if there are direct entries, then take them to generate the matrix otherwise use the pw judgments
                    dir_count=0
                    dir_vector=reqLib.np.empty([size], dtype = float)
                    for dir_val in range(size):
                        dir_item=array_excel.iat[dir_val+row,size+1]
                        if(dir_item>0):
                            dir_vector[dir_count]=dir_item
                            dir_count+=1
                            if(verbal):
                                print("-->dir-->",dir_item)
                    if(dir_count==size):
                        if(verbal):
                            print("will not read judgments")
                        pc_matrix=convDirect2PairwiseMatrix(dir_vector,verbal)  
                    else:
                        for a in range(size):
                            for b in range (size):
                                item=array_excel.iat[a+row,b+1]
                                # print("item",item)
                                if a==b:
                                    pc_matrix[a,b]=1.0
                                elif a<b:
                                    pc_matrix[a,b]=item
                                    if(item!=0):
                                        pc_matrix[b,a]=1./item
                                    else:
                                        pc_matrix[b,a]=0
                    if(verbal==True): print(pc_matrix)
                    model.all_pc_matrices.append(pc_matrix)
                    row=row+size+1
                row+=1

def importRatingsFromExcel(model,filepath,sheet_name, verbal=False):

    all_excel = reqLib.pd.read_excel (filepath, sheet_name=sheet_name, header=None) 
    array_excel=all_excel.fillna(0)
    file_len=len(array_excel)
    # if(verbal==True):
    #     print("File Len=", file_len)
    #     print("excel",array_excel)
    row=0
    if(model.modelType=="ratings"):
        while(row<file_len):
            my_crit_name=array_excel.iat[row,0]
            my_crit=model.rateModel.getRatCritObjByName(my_crit_name)
            
            if (verbal):
                print(f"row: {row}--for criterion {my_crit_name}")

            my_scale=model.rateModel.getScaleOfCriterion(my_crit)
            if (verbal):
                print(f"scale {my_scale.name}")
            row+=2
            size=len(my_scale.members)
            
            # for each criterion read a members by members matrix and save it for calculations
            
            pc_matrix =  reqLib.np.empty([size, size], dtype = float)
            #check if there are direct entries, then take them to generate the matrix otherwise use the pw judgments
            dir_count=0
            dir_vector=reqLib.np.empty([size], dtype = float)
            for dir_val in range(size):
                dir_item=array_excel.iat[dir_val+row,size+1]
                if(dir_item>0):
                    dir_vector[dir_count]=dir_item
                    dir_count+=1
                    if(verbal):
                        print("-->dir-->",dir_item)
            if(dir_count==size):
                if(verbal):
                    print("will not read judgments")
                pc_matrix=convDirect2PairwiseMatrix(dir_vector,verbal)  
            else:    
                for a in range(size):
                    for b in range (size):
                        item=array_excel.iat[a+row,b+1]
                        # print("item",item)
                        if a==b:
                            pc_matrix[a,b]=1.0
                        elif a<b:
                            pc_matrix[a,b]=item
                            if(item!=0):
                                pc_matrix[b,a]=1./item
                            else:
                                pc_matrix[b,a]=0
            if(verbal==True):
                print(pc_matrix)
          
            my_val_mat=reqLib.cp.deepcopy(pc_matrix)
            model.rateModel.updateScaleMatrix(my_scale.name,my_val_mat.copy())
            row+=size+3       
    else:
        print("Cannot read ratings if model is not set up as a ratings model")
    
def importRatTableFromExcel(model,filepath,sheet_name, verbal=False):

    all_excel = reqLib.pd.read_excel (filepath, sheet_name=sheet_name, header=None) 
    array_excel=all_excel.fillna(0)
    file_len=len(array_excel)
    # if(verbal==True):
    #     print("File Len=", file_len)
    #     print("excel",array_excel)
    row=1
    n_alt=len(model.rateModel.ratAlternatives)
    n_crit=len(model.rateModel.ratCriteria)
    r_matrix = reqLib.np.full((n_alt, n_crit), -1)
    if(model.modelType=="ratings"):
        for i,alt in enumerate(model.rateModel.ratAlternatives):
            col=0
            for j,crit in enumerate(model.rateModel.ratCriteria):
                value=array_excel.iat[row+i,col+j+1]
                scale=model.rateModel.getScaleOfCriterion(crit)
                
                priority=scale.getmValueBymName(value,verbal)
                index=scale.getmIndexBymName(value,verbal)

                r_matrix[i,j]=index

                if(verbal==True):
                    print(f"Reading for crit: {crit.name} and alt: {alt.name}------{value}------>{scale.name} ---{priority} ---> at loc {index}")
            if(verbal==True):
                print("------\n")
        
        model.rateModel.ratMatrix=reqLib.cp.deepcopy(r_matrix)
        if(verbal==True):
            print(reqLib.tabulate(r_matrix.tolist(), tablefmt="grid"))
        
def calcRatResults(model,exportfile,verbal=False):
    cdf_calc.calcRateCritV(model,verbal=verbal)
    export4ExcelRatingScales(model,exportfile,True,False,verb=verbal)
    cdf_calc.appendScaleResults2Excel(model,exportfile,"rating_scales", normalbar=False,idealbar=True,verb=verbal)
    export4ExcelRatingsTable(model,exportfile,verbal)
    cdf_calc.appendRateTblResults2Excel(model,exportfile,"rating_table", normalbar=False,idealbar=True,verb=verbal)
def convDirect2PairwiseMatrix(vector,verbal=False):
    size=len(vector)
    pc_matrix =  reqLib.np.empty([size, size], dtype = float)
    for i in range(0,size):
        for j in range(0,size):
            pc_matrix[i,j]=vector[i]/vector[j]
            # print("\n",i,vector[i],j,vector[j],pc_matrix[i,j])
    if(verbal==True):
        print("Vector:",vector,"\nPC Matrix:\n",pc_matrix)
    return pc_matrix    

def readRatScaleRPCfile(scaleName,fileName,verbal=False):
    # read a superdecisions generated scale file and return the vector of member names and judgment pc matrix
    with open(fileName, 'r') as file:
        file_contents = file.read()
    raw_lines = file_contents.split('\n')
    file.close()
    names=[]
    judgments=[]
    values=[]
    memberText="ratings newCategory -network $net -criteria $crit -group $group -category {"
    valueText="set values {"
    for line in raw_lines:
        if( memberText in line):
            info = line.split(memberText)[1].strip("}")
            names.append(info)
        if( valueText in line):
            info = line.split(valueText)[1].strip("}")
            judgments=info.split(",")[0]
            input_list = judgments.split()

    # Convert each element to a float using a list comprehension
    fl_judg_list = [float(elem) for elem in input_list]
    nnodes=len(names)
    matrix_judg=cdf_calc.vector2matrix(fl_judg_list,nnodes)
    priorities=cdf_calc.priorityVector(matrix_judg)
    
    indx=0
    for name in names:
        values.append([name,priorities[indx]])
        indx+=1
    scale=cdf_rat.RatScale(scaleName)
    scale.defineScaleByValue(matrix_judg,False,*values)

    if(verbal):
        print("Reading Judgments from file:"+fileName)
        print(names)
        print(fl_judg_list)
        print("nodes:",nnodes)
        print(matrix_judg)
        print(priorities)
    return scale

def readSDMODfile(modelName,fileName,export=False,verbal=False):
    new_model=cdf_str.Model(modelName)

    with open(fileName, 'r') as file:
        file_contents = file.read()
    raw_lines = file_contents.split('\n')
    file.close()
    netText="set net ["
    structureText="c-network readAllCompares -network $net -source {"
    structureStart=0
    cntNd=0
    totNd=0
    cntCl=0
    clusters=[]
    nodes=[]
    node_pwc_order=[]
    temp_all_pc_matrices=[]
    if (export):
        f_op=open(modelName+'.py', 'w')
        f_op.write('from AhpAnpLib import inputs_AHPLib as input\nfrom AhpAnpLib import structs_AHPLib as str\nfrom AhpAnpLib import calcs_AHPLib as calc\nfrom AhpAnpLib import ratings_AHPLib as rate\nfrom AhpAnpLib import functions_AHPLib as reqLib\n')
        f_op.write('\nnew_model=str.Model("'+modelName+'")\n')
    #find starting point with interesting data
    for i,line in enumerate(raw_lines):
        if( netText in line):
            #tbd create multiple model objects for multilevel nets
            break
        else:
            if(structureText in line):
                structureStart=i+1
                cntCl=int(raw_lines[structureStart])
            line=structureStart+1
    #read model's  clusters and nodes
    for i in range(0,cntCl):
        component=raw_lines[line]
        line+=1
        clusters.append(component) 
        cl=cdf_str.Cluster(component,i)
        if(export):
            f_op.write('\ncl'+str(i)+'=str.Cluster("'+component+'",'+str(i)+')')
        cntNd=int(raw_lines[line])
        line+=1

        for j in range(0,cntNd):
            component=raw_lines[line]
            line+=1
            nodes.append(component)
            nd=cdf_str.Node(component,j+totNd)
            cl.addNode2Cluster(nd)
            if(export):
                f_op.write('\nnd'+str(i)+'=str.Node("'+component+'",'+str(j+totNd)+')')
                f_op.write('\ncl'+str(i)+'.addNode2Cluster(nd'+str(i)+')')
        totNd+=cntNd
        new_model.addCluster2Model(cl)
        if(export):
            f_op.write('\nnew_model.addCluster2Model(cl'+str(i)+')\n')
    #now add cluster info
    line+=1
    component=raw_lines[line]
    info=reqLib.re.split(r'[,\s]+', component)
    int_info_list = [int(elem) for elem in info if elem.strip()]
    # print(int_info_list)
    matSize=int_info_list[0]
    del int_info_list[0]

    cl_neighbors_labels=[]
    for i,item in enumerate(int_info_list):
        if((i%2)==0):
            cur_cl=clusters[item]
        else:
            cl_neighbors_labels.insert(item,cur_cl)
    line+=1
    cl_neigh=[]
    for clLine in range(0,matSize):
        component=raw_lines[line]
        judgments=component.split(" ")
        cl_neighbors_list = [float(elem) for elem in judgments if elem.strip()]
        cl_neigh.append(cl_neighbors_list)
        line+=1
        
    cl_neighbors_df = reqLib.pd.DataFrame(cl_neigh,cl_neighbors_labels,cl_neighbors_labels)  
    new_model.cluster_neighbors_mtrx=cl_neighbors_df

    #for each cluster read cluster matrix
    # print(cntCl)
    for clMatCnt in range(cntCl):
        cl_df,cur_id,line=sdmodLines2PwMatrix(new_model,raw_lines,line,f_op,verbal)
        #nodes section
    #find how many node pwc matrices you will need to read
    component=raw_lines[line]
    temp_ln=line
    count_pw=0
    while component!="}":
        if(component==""):
            count_pw+=1
        # print(component)
        temp_ln+=1
        component=raw_lines[temp_ln]
    
    # print(count_pw)
    #import node pairwise comparisons
    for ndMatCount in range(count_pw):
        nd_df,cur_id,line=sdmodLines2PwMatrix(new_model,raw_lines,line,f_op,verbal)
        if(nd_df.empty==False):
            temp_all_pc_matrices.append(nd_df.to_numpy())
            node_pwc_order.append(cur_id)
    # print(node_pwc_order)
    #re-order to match internal order
    new_model.all_pc_matrices=reqLib.cp.deepcopy(reorder_pwc_np(new_model,temp_all_pc_matrices,node_pwc_order))
    # print("2nd round")
    
    input_file=new_model.name+"_empty.xlsx"
    input_file_full=new_model.name+"_filledIn.xlsx"
    output_file=new_model.name+"_results.xlsx"

    export4ExcelQuestFull(new_model,input_file,show_estimates=False,verb=False)
    if (verbal):
        print("\npc matrices------- ")
        print(new_model.all_pc_matrices)
    cdf_calc.calcAHPMatricesSave2File(new_model,input_file,output_file,inputFileUse=False,normalbar=False,idealbar=True,verbal=False)
    #add the inputs to an input file to easily modify
    cdf_calc.copyExcelSheet(output_file, input_file_full, "pairwise_comp",False)
    # save pairwise comparison matrices 

    if(verbal):
        # print(raw_lines)
        # print(structureStart)
        print(cntCl)
        print(clusters)
        print(nodes)
        print(cl_neighbors_df)
        new_model.printStruct()
    if(export):
        f_op.write('\nnew_model.printStruct()')
        f_op.write('\ninput.export4ExcelQuestFull(new_model,"py_file_'+input_file+'",verb=False)')
        f_op.write('\ncalc.calcAHPMatricesSave2File(new_model,"'+input_file_full+'","py_file_'+output_file+'",inputFileUse=True,normalbar=False,idealbar=True,verbal=False)')
        f_op.write('\ncalc.copyExcelSheet("py_file_'+output_file+'","py_file_'+input_file_full+'", "pairwise_comp",False)')
        f_op.close()    
    return new_model

def convertJudgVec2Pairwise(my_list,my_order,mat_size):
    count=0
    print("List",my_list)
    print("Order",my_order)
    my_df=reqLib.np.ones([mat_size, mat_size], dtype = float)
    for i in range(mat_size):
        for j in range(mat_size):
            if(i<j):
                my_df[i,j]=my_list[count]
                if(my_list[count]!=0):
                    my_df[j,i]=1./my_df[i,j]
                count+=1       
    return my_df

def sdmodLines2PwMatrix(my_model,raw_lines,start_ln,file_op=None,verbal=False):
    cur_id=[]
    line=start_ln
    matID=raw_lines[line]
    if(matID.isdigit()):
        my_cluster=my_model.clusters[int(matID)]
        cur_id=[my_cluster]
        if(verbal):
            print("reading cluster with ID"+str(matID))
    else:
        
        info=reqLib.re.split(r'[,\s]+', matID)
        int_conn_list = [int(elem) for elem in info if elem.strip()]
        par_cluster=my_model.clusters[int_conn_list[0]]
        my_node=my_model.clusters[int_conn_list[0]].nodes[int_conn_list[1]]
        con_cluster=my_model.clusters[int_conn_list[2]]
        
    line+= 1
    #tbd the ids in sdmod are not overall are with respect to the node
    component=raw_lines[line]
    info=reqLib.re.split(r'[,\s]+', component)
    int_cl_list = [int(elem) for elem in info if elem.strip()]


    clMatSize=int_cl_list[0]
    if(verbal): print(clMatSize)
    if(clMatSize>0 and matID.isdigit()==False):
        cur_id=[par_cluster,my_node,con_cluster]
        if(verbal):
                print(f"Node:{my_node.name} of cluster {par_cluster.name} is connected to cluster {con_cluster.name}")

    del int_cl_list[0]
    #order of clusters in matrix
    cl_labels=[]
    if (clMatSize>0):
        for i,item in enumerate(int_cl_list):
            if((i%2)==0):
                if(matID.isdigit()):
                    cur=my_model.clusters[item]
                else:
                    # print(con_cluster)
                    # print(item)
                    
                    cur=con_cluster.nodes[item]
                    #add connection my_node to cur
                    my_model.addNodeConnectionFromTo(my_node.name,cur.name,False)
                    if(file_op!=None):
                        file_op.write('\nnew_model.addNodeConnectionFromTo("'+my_node.name+'","'+cur.name+'",False)')
            else:
                cl_labels.insert(item,cur.name)
    # print(cl_labels)
    #this will give me the connection...check if nodes or clusters if cluster the cluster id but if node clA,node of A, clB
    if (clMatSize>1):
        cl_j=[]
        for item in range(clMatSize):
            line+=1
            component=raw_lines[line]
            info=reqLib.re.split(r'[,\s]+', component)
            cl_j_list = [float(elem) for elem in info if elem.strip()]
            # print(cl_j_list)
            cl_j+=cl_j_list
        line+=1
        pc_j_matrix =convertJudgVec2Pairwise(cl_j,int_cl_list,clMatSize)
        # cl_df = reqLib.pd.DataFrame(pc_j_matrix,cl_labels,cl_labels)
        cl_df = reqLib.pd.DataFrame(pc_j_matrix)
        if(verbal):
            print(cl_df)
    else:
        cl_df = reqLib.pd.DataFrame([])
        # cl_df=None
        line+=2
    return cl_df,cur_id,line

def reorder_pwc_np(model,np_array,list_conn):
    cnt=0
    new_array = []
    for cluster in sorted( model.clusters, key=reqLib.op.attrgetter('order')):
            for nodeFrom in sorted(cluster.nodes, key=reqLib.op.attrgetter('order')): 
                connectedClusters=model.retAllClusterConnectionsFromNode(nodeFrom.name)
                for clusterPWC in sorted(connectedClusters, key=reqLib.op.attrgetter('order')):
                    
                    # print(f"Expecting CL A:{cluster.name} - ND: {nodeFrom.name} - CL B: {clusterPWC.name}")
                    # print(f"I have CL A:{list_conn[cnt][0].name} - ND: {list_conn[cnt][1].name} - CL B: {list_conn[cnt][2].name}") 
                    pos=find_index_pwc(list_conn,cluster,nodeFrom,clusterPWC)
                    # print(pos)
                    new_array.insert(cnt,np_array[pos])
                    # print(new_array)
                    cnt+=1
    return new_array

def find_index_pwc(lst, cluster_a, node_a, cluster_b):
    for i, sublst in enumerate(lst):
        if sublst == [cluster_a, node_a, cluster_b]:
            # print(f"I found {cluster_a} -{node_a} -{cluster_b} in {i}\n")
            return i
    return -1

def convIndex2ExcelRef(row,col):
    col_letter = reqLib.get_column_letter(col)
    excel_ref=col_letter+str(row)
    # print(f"c={col}-->letter{col_letter}")
    return excel_ref

def convExcelRef2Index(excel_ref):
    col = ''.join([i for i in excel_ref if not i.isdigit()])
    row = int(''.join([i for i in excel_ref if i.isdigit()]))
    return row,col

def set_cell_value_and_style(cell, value, fill=None, border=None, font_color=None, font_type=False):
    cell.value = value
    if fill:
        cell.fill = fill
    if border:
        cell.border = border
    if font_color:
        cell.font=reqLib.Font(color=font_color, bold=font_type)
        
def addDropDown2ExcelCell(worksheet,c_from,c_to,worksheet_title,cell_to_validate): 

   # Set up the data validation for the desired cell
   dv = reqLib.DataValidation(type="list", formula1=f'={worksheet_title}!{c_from}:{c_to}')
   worksheet.add_data_validation(dv)
   dv.add(worksheet[cell_to_validate])

   return worksheet

